/*-
 * Copyright (C)2021 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ident "@(#) Copyright (C)2021 @BABOLO http://www.babolo.ru/"
#ident "@(#) $Id: pam.c,v 1.2 2021/03/14 01:43:02 babolo Exp $"

#define BLIN_COMPAT     3
#define MIFE_COMPAT     5
#define Bpars_COMPAT    4
#define RECOBE_COMPAT   4

#include <sys/types.h>
#include <sys/event.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sysexits.h>
#include <strings.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <stdio.h>
#include <errno.h>
#include <zlib.h>
#include <babolo/BLINflag.h>
#include <babolo/parser.h>
#include <babolo/recobe.h>
#include <mife.h>
#include "pngtr.h"

int
/*****************************************************************************************************
 *****************************************************************************************************
 ****    fd  cf->num    f                                           ****
 *****************************************************************************************************
                                                                                                  ****/
pngtr_outpam(pngtr_conf *cf, int fd) {                                                           /****
                                                                                                  ****
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags (cf->flags)
    pngtr_syncolor    syncolor;
    pngtr_syncolor    smcolor;
    pngtr_operand    *palette  = NULL;
    pngtr_convey     *convey;
    u_int16_t         acolor;
    pngtr_operand    *alfa     = NULL;
    u_int16_t         grey;
    int               hlen;
    char             *buf;
    pngtr_syncolor   *img;
    int               ex    = EX_OK;
#   define O(R, C) img[convey->IHDR.width * r + c]

    ifBLIN_QX3("+ %08X %u", cf->flags, cf->num);
    PNGTR_X((cf->num >= PNGTR_CONVEYNUM),            ("Ill convey num %d", cf->num), ERANGE, EX_USAGE);
    convey = &cf->convey[cf->num];
    for (u_int chctr = 0; chctr < convey->chunum; ++chctr) {
        if  (PNGTR_tRNS == convey->chunks[chctr].type) {
            alfa = &convey->chunks[chctr].operands[0];
            break;
    }   }
    hlen = asprintf( &buf
                   , "P7\nWIDTH %u\nHEIGHT %u\nDEPTH 4\nMAXVAL 65535\nTUPLTYPE RGB_ALPHA\nENDHDR\n"
                   , convey->IHDR.width
                   , convey->IHDR.height
                   );
    PNGTR_W( (hlen != mife_writ(fd, buf, hlen))
           , ("mife_writ %s", !convey->png.nm ? "" : convey->png.nm)
           , EX_IOERR
           );
    PNGTR_W( !(img = calloc(convey->IHDR.width * convey->IHDR.height, sizeof(pngtr_syncolor)))
           , ("calloc")
           , EX_OSERR
           );
    if  (convey->IHDR.palette) {
        for (u_int i = 0; i < convey->chunum; ++i) {
            if  (PNGTR_PLTE == convey->chunks[i].type) {
                palette = convey->chunks[i].operands;
                break;
    }   }   }
    switch(convey->IHDR.depth) {
#   define I(R, C, N) convey->img.cont[((R) * convey->IHDR.width + (N) - 1) / (N) + (C) / (N)]
    case 1:
        switch(convey->IHDR.colortype) {
        case 0:
            if  (!!alfa && (0 < alfa->maxindex)) acolor = (u_int16_t)recobe_load(alfa->val, -2);
            for (u_int r = 0; r < convey->IHDR.height; ++r) {
                for (u_int c = 0; c < convey->IHDR.width; ++c) {
                    grey = (I(r, c, 8) >> (7 - (c & 7))) & 1;
                    if  (!!alfa && (0 < alfa->maxindex) && (grey == acolor)) {
                        O(r, c).a = 0;
                    } else {
                        O(r, c).a = 65535;
                    }
                    if  (!!grey) grey = 65535;
                    O(r, c).r = O(r, c).g = O(r, c).b = grey;
            }   }
            break;
        case 3:
            PNGTR_X( !palette
                   , ("Palette absent %s", !convey->png.nm ? "" : convey->png.nm)
                   , EX_DATAERR
                   , EFTYPE
                   );
            for (u_int r = 0; r < convey->IHDR.height; ++r) {
                for (u_int c = 0; c < convey->IHDR.width; ++c) {
                    grey = (I(r, c, 8) >> (7 - (c & 7))) & 1;
                    if  (grey >= palette->maxindex) {
                        ifBLIN_QX0( "Palette maxindex %u <= %u in %s"
                                  , palette->maxindex
                                  , grey
                                  , !convey->png.nm ? "" : convey->png.nm
                                  );
                    } else {
                        O(r, c).r = ((pngtr_color*)palette->val)[grey].r;
                        O(r, c).g = ((pngtr_color*)palette->val)[grey].g;
                        O(r, c).b = ((pngtr_color*)palette->val)[grey].b;
                        if  (!!alfa && (grey < alfa->maxindex)) {
                            O(r, c).a = alfa->val[grey];
                            O(r, c).a |= O(r, c).a << 8;
                        } else {
                            O(r, c).a = 65535;
                        }
                        O(r, c).r |= O(r, c).r << 8;
                        O(r, c).g |= O(r, c).g << 8;
                        O(r, c).b |= O(r, c).b << 8;
            }   }   }
            break;
        default:
            ifBLIN_QX0( "colortype %u depth %u in %s"
                      , convey->IHDR.colortype
                      , convey->IHDR.depth
                      , !convey->png.nm ? "" : convey->png.nm
                      );
            ex = EX_USAGE;
            errno = EINVAL;
            goto out;
        }
        break;
    case 2:
        switch(convey->IHDR.colortype) {
        case 0:
            if  (!!alfa && (0 < alfa->maxindex)) acolor = (u_int16_t)recobe_load(alfa->val, -2);
            for (u_int r = 0; r < convey->IHDR.height; ++r) {
                for (u_int c = 0; c < convey->IHDR.width; ++c) {
                    grey = (I(r, c, 4) >> (6 - 2 * (c & 3))) & 3;
                    if  (!!alfa && (0 < alfa->maxindex) && (grey == acolor)) {
                        O(r, c).a = 0;
                    } else {
                        O(r, c).a = 65535;
                    }
                    grey = (u_int16_t[]){0x0000, 0x5555, 0xAAAA, 0xFFFF}[grey];
                    O(r, c).r = O(r, c).g = O(r, c).b = grey;
            }   }
            break;
        case 3:
            PNGTR_X( !palette
                   , ("Palette absent %s", !convey->png.nm ? "" : convey->png.nm)
                   , EX_DATAERR
                   , EFTYPE
                   );
            for (u_int r = 0; r < convey->IHDR.height; ++r) {
                for (u_int c = 0; c < convey->IHDR.width; ++c) {
                    grey = (I(r, c, 4) >> (6 - 2 * (c & 3))) & 3;
                    if  (grey >= palette->maxindex) {
                        ifBLIN_QX0( "Palette maxindex %u <= %u in %s"
                                  , palette->maxindex
                                  , grey
                                  , !convey->png.nm ? "" : convey->png.nm
                                  );
                    } else {
                        if  (!!alfa && (grey < alfa->maxindex)) {
                            O(r, c).a = alfa->val[grey];
                            O(r, c).a |= O(r, c).a << 8;
                        } else {
                            O(r, c).a = 65535;
                        }
                        O(r, c).r = ((pngtr_color*)palette->val)[grey].r;
                        O(r, c).g = ((pngtr_color*)palette->val)[grey].g;
                        O(r, c).b = ((pngtr_color*)palette->val)[grey].b;
                        O(r, c).r |= O(r, c).r << 8;
                        O(r, c).g |= O(r, c).g << 8;
                        O(r, c).b |= O(r, c).b << 8;
            }   }   }
            break;
        default:
            ifBLIN_QX0( "colortype %u depth %u in %s"
                      , convey->IHDR.colortype
                      , convey->IHDR.depth
                      , !convey->png.nm ? "" : convey->png.nm
                      );
            ex = EX_USAGE;
            errno = EINVAL;
            goto out;
        }
        break;
    case 4:
        switch(convey->IHDR.colortype) {
        case 0:
            if  (!!alfa && (0 < alfa->maxindex)) acolor = (u_int16_t)recobe_load(alfa->val, -2);
            for (u_int r = 0; r < convey->IHDR.height; ++r) {
                for (u_int c = 0; c < convey->IHDR.width; ++c) {
                    grey = (I(r, c, 2) >> (4 - 4 * (c & 1))) & 15;
                    if  (!!alfa && (0 < alfa->maxindex) && (grey == acolor)) {
                        O(r, c).a = 0;
                    } else {
                        O(r, c).a = 65535;
                    }
                    grey = (u_int16_t[]){0x0000, 0x1111, 0x2222, 0x3333, 0x4444, 0x5555, 0x6666, 0x7777, 0x8888, 0x9999, 0xAAAA, 0xBBBB, 0xCCCC, 0xDDDD, 0xEEEE, 0xFFFF}[grey];
                    O(r, c).r = O(r, c).g = O(r, c).b = grey;
            }   }
            break;
        case 3:
            PNGTR_X( !palette
                   , ("Palette absent %s", !convey->png.nm ? "" : convey->png.nm)
                   , EX_DATAERR
                   , EFTYPE
                   );
            for (u_int r = 0; r < convey->IHDR.height; ++r) {
                for (u_int c = 0; c < convey->IHDR.width; ++c) {
                    grey = (I(r, c, 2) >> (4 - 4 * (c & 1))) & 15;
                    if  (grey >= palette->maxindex) {
                        ifBLIN_QX0( "Palette maxindex %u <= %u in %s"
                                  , palette->maxindex
                                  , grey
                                  , !convey->png.nm ? "" : convey->png.nm
                                  );
                    } else {
                        if  (!!alfa && (grey < alfa->maxindex)) {
                            O(r, c).a = alfa->val[grey];
                            O(r, c).a |= O(r, c).a << 8;
                        } else {
                            O(r, c).a = 65535;
                        }
                        O(r, c).r = ((pngtr_color*)palette->val)[grey].r;
                        O(r, c).g = ((pngtr_color*)palette->val)[grey].g;
                        O(r, c).b = ((pngtr_color*)palette->val)[grey].b;
                        O(r, c).r |= O(r, c).r << 8;
                        O(r, c).g |= O(r, c).g << 8;
                        O(r, c).b |= O(r, c).b << 8;
            }   }   }
            break;
        default:
            ifBLIN_QX0( "colortype %u depth %u in %s"
                      , convey->IHDR.colortype
                      , convey->IHDR.depth
                      , !convey->png.nm ? "" : convey->png.nm
                      );
            ex = EX_USAGE;
            errno = EINVAL;
            goto out;
        }
        break;
    case 8:
        switch(convey->IHDR.colortype) {
#   undef  I
#   define X(R, C, N, M) ((N) * ((R) * convey->IHDR.width + (C)) + (M))
#   define I(R, C, N, M) convey->img.cont[X((R), (C), (N), (M))]
        case 0:
            if  (!!alfa && (0 < alfa->maxindex)) acolor = (u_int16_t)recobe_load(alfa->val, -2);
            for (u_int r = 0; r < convey->IHDR.height; ++r) {
                for (u_int c = 0; c < convey->IHDR.width; ++c) {
                    grey = I(r, c, 1, 0) & 0x00FF;
                    if  (!!alfa && (0 < alfa->maxindex) && (grey == acolor)) {
                        O(r, c).a = 0;
                    } else {
                        O(r, c).a = 65535;
                    }
                    grey |= grey << 8;
                    O(r, c).r = O(r, c).g = O(r, c).b = grey;
            }   }
            break;
        case 2:
            if  (!!alfa && (0 < alfa->maxindex)) {
                syncolor.r = (u_int16_t)recobe_load(((pngtr_color2 *)alfa->val)->r, -2);
                syncolor.g = (u_int16_t)recobe_load(((pngtr_color2 *)alfa->val)->g, -2);
                syncolor.b = (u_int16_t)recobe_load(((pngtr_color2 *)alfa->val)->b, -2);
            }
            for (u_int r = 0; r < convey->IHDR.height; ++r) {
                for (u_int c = 0; c < convey->IHDR.width; ++c) {
                    O(r, c).r = I(r, c, 3, 0) & 0x00FF;
                    O(r, c).g = I(r, c, 3, 1) & 0x00FF;
                    O(r, c).b = I(r, c, 3, 2) & 0x00FF;
                    if  (  !!alfa
                        && (0 < alfa->maxindex)
                        && (O(r, c).r == syncolor.r)
                        && (O(r, c).g == syncolor.g)
                        && (O(r, c).b == syncolor.b)
                        ) {
                        O(r, c).a = 0;
                    } else {
                        O(r, c).a = 65535;
                    }
                    O(r, c).r |= O(r, c).r << 8;
                    O(r, c).g |= O(r, c).g << 8;
                    O(r, c).b |= O(r, c).b << 8;
            }   }
            break;
        case 3:
            PNGTR_X( !palette
                   , ("Palette absent %s", !convey->png.nm ? "" : convey->png.nm)
                   , EX_DATAERR
                   , EFTYPE
                   );
            for (u_int r = 0; r < convey->IHDR.height; ++r) {
                for (u_int c = 0; c < convey->IHDR.width; ++c) {
                    grey = I(r, c, 1, 0) & 0x00FF;
                    if  (grey >= palette->maxindex) {
                        ifBLIN_QX0( "Palette maxindex %u <= %u in %s"
                                  , palette->maxindex
                                  , grey
                                  , !convey->png.nm ? "" : convey->png.nm
                                  );
                    } else {
                        if  (!!alfa && (grey < alfa->maxindex)) {
                            O(r, c).a = alfa->val[grey];
                            O(r, c).a |= O(r, c).a << 8;
                        } else {
                            O(r, c).a = 65535;
                        }
                        O(r, c).r = ((pngtr_color*)palette->val)[grey].r;
                        O(r, c).g = ((pngtr_color*)palette->val)[grey].g;
                        O(r, c).b = ((pngtr_color*)palette->val)[grey].b;
                        O(r, c).r |= O(r, c).r << 8;
                        O(r, c).g |= O(r, c).g << 8;
                        O(r, c).b |= O(r, c).b << 8;
            }   }   }
            break;
        case 4:
            for (u_int r = 0; r < convey->IHDR.height; ++r) {
                for (u_int c = 0; c < convey->IHDR.width; ++c) {
                    grey = I(r, c, 2, 0) & 0x00FF;
                    grey |= grey << 8;
                    O(r, c).r = O(r, c).g = O(r, c).b = grey;
                    O(r, c).a = I(r, c, 2, 1) & 0x00FF;
                    O(r, c).a |= O(r, c).a << 8;
            }   }
            break;
        case 6:
            for (u_int r = 0; r < convey->IHDR.height; ++r) {
                for (u_int c = 0; c < convey->IHDR.width; ++c) {
                    O(r, c).r = I(r, c, 4, 0) & 0x00FF;
                    O(r, c).g = I(r, c, 4, 1) & 0x00FF;
                    O(r, c).b = I(r, c, 4, 2) & 0x00FF;
                    O(r, c).a = I(r, c, 4, 3) & 0x00FF;
                    O(r, c).r |= O(r, c).r << 8;
                    O(r, c).g |= O(r, c).g << 8;
                    O(r, c).b |= O(r, c).b << 8;
                    O(r, c).a |= O(r, c).a << 8;
            }   }
            break;
        default:
            ifBLIN_QX0( "colortype %u depth %u in %s"
                      , convey->IHDR.colortype
                      , convey->IHDR.depth
                      , !convey->png.nm ? "" : convey->png.nm
                      );
            ex = EX_USAGE;
            errno = EINVAL;
            goto out;
        }
        break;
#   undef  I
#   define I(R, C, N, M) (u_int16_t)recobe_load(&convey->img.cont[2 * X((R), (C), (N), (M))], -2)
    case 16:
        switch(convey->IHDR.colortype) {
        case 0:
            if  (!!alfa && (0 < alfa->maxindex)) acolor = (u_int16_t)recobe_load(alfa->val, -2);
            for (u_int r = 0; r < convey->IHDR.height; ++r) {
                for (u_int c = 0; c < convey->IHDR.width; ++c) {
                    grey = I(r, c, 1, 0);
                    if  (!!alfa && (0 < alfa->maxindex) && (grey == acolor)) {
                        O(r, c).a = 0;
                    } else {
                        O(r, c).a = 65535;
                    }
                    recobe_store((u_char *)&O(r, c).r, -2, grey);
                    recobe_store((u_char *)&O(r, c).g, -2, grey);
                    recobe_store((u_char *)&O(r, c).b, -2, grey);
            }   }
            break;
        case 2:
            if  (!!alfa && (0 < alfa->maxindex)) {
                syncolor.r = (u_int16_t)recobe_load(((pngtr_color2 *)alfa->val)->r, -2);
                syncolor.g = (u_int16_t)recobe_load(((pngtr_color2 *)alfa->val)->g, -2);
                syncolor.b = (u_int16_t)recobe_load(((pngtr_color2 *)alfa->val)->b, -2);
            }
            for (u_int r = 0; r < convey->IHDR.height; ++r) {
                for (u_int c = 0; c < convey->IHDR.width; ++c) {
                    smcolor.r = I(r, c, 3, 0);
                    smcolor.g = I(r, c, 3, 1);
                    smcolor.b = I(r, c, 3, 2);
                    if  (  !!alfa
                        && (0 < alfa->maxindex)
                        && (smcolor.r == syncolor.r)
                        && (smcolor.g == syncolor.g)
                        && (smcolor.b == syncolor.b)
                        ) {
                        O(r, c).a = 0;
                    } else {
                        O(r, c).a = 65535;
                    }
                    recobe_store((u_char *)&O(r, c).r, -2, smcolor.r);
                    recobe_store((u_char *)&O(r, c).g, -2, smcolor.g);
                    recobe_store((u_char *)&O(r, c).b, -2, smcolor.b);
            }   }
            break;
        case 4:
            for (u_int r = 0; r < convey->IHDR.height; ++r) {
                for (u_int c = 0; c < convey->IHDR.width; ++c) {
                    grey = I(r, c, 2, 0);
                    recobe_store((u_char *)&O(r, c).r, -2, grey);
                    recobe_store((u_char *)&O(r, c).g, -2, grey);
                    recobe_store((u_char *)&O(r, c).b, -2, grey);
                    recobe_store((u_char *)&O(r, c).a, -2, I(r, c, 2, 1));
            }   }
            break;
        case 6:
            for (u_int r = 0; r < convey->IHDR.height; ++r) {
                for (u_int c = 0; c < convey->IHDR.width; ++c) {
                    recobe_store((u_char *)&O(r, c).r, -2, I(r, c, 4, 0));
                    recobe_store((u_char *)&O(r, c).g, -2, I(r, c, 4, 1));
                    recobe_store((u_char *)&O(r, c).b, -2, I(r, c, 4, 2));
                    recobe_store((u_char *)&O(r, c).a, -2, I(r, c, 4, 3));
            }   }
            break;
        default:
            ifBLIN_QX0( "colortype %u depth %u in %s"
                      , convey->IHDR.colortype
                      , convey->IHDR.depth
                      , !convey->png.nm ? "" : convey->png.nm
                      );
            ex = EX_USAGE;
            errno = EINVAL;
            goto out;
        }
        break;
    default:
        ifBLIN_QX0( "depth %u in %s"
                  , convey->IHDR.depth
                  , !convey->png.nm ? "" : convey->png.nm
                  );
        ex = EX_USAGE;
        errno = EINVAL;
        goto out;
    }
    PNGTR_W( (convey->IHDR.width * convey->IHDR.height * sizeof(pngtr_syncolor) != mife_writ(fd, img, convey->IHDR.width * convey->IHDR.height * sizeof(pngtr_syncolor)))
           , ("mife_writ %s", !convey->png.nm ? "" : convey->png.nm)
           , EX_IOERR
           );
out:
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef  I
#   undef  O
#   undef  blin_internal_flags
}
