/*-
 * Copyright (C)2021 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: pngtr.h,v 1.33 2021/03/14 18:26:47 babolo Exp $
 */

#define PNGTR_CHUNKSZ   32768
#define PNGTR_CONVEYNUM 16
#define PNGTR_SIGNATURE {0x89, 0x50, 0x4E, 0x47, 0x0D, 0x0A, 0x1A, 0x0A}

/***************************************************************************************************** 
 ****                                                                                   **** 
 *****************************************************************************************************/
#define PNGTR_IHDR   1
#define PNGTR_PLTE   2
#define PNGTR_IDAT   3
#define PNGTR_IEND   4
#define PNGTR_cHRM   5
#define PNGTR_gAMA   6
#define PNGTR_iCCP   7
#define PNGTR_sBIT   8
#define PNGTR_sRGB   9
#define PNGTR_bKGD  10
#define PNGTR_hIST  11
#define PNGTR_tRNS  12
#define PNGTR_pHYs  13
#define PNGTR_sPLT  14
#define PNGTR_tIME  15
#define PNGTR_iTXt  16
#define PNGTR_tEXt  17
#define PNGTR_zTXt  18
#define PNGTR_oFFs  19
#define PNGTR_pCAL  20
#define PNGTR_sCAL  21
#define PNGTR_gIFg  22
#define PNGTR_gIFt  23
#define PNGTR_gIFx  24
#define PNGTR_sTER  25
#define PNGTR_dSIG  26
#define PNGTR_eXIf  27
#define PNGTR_fRAc  28
#define PNGTR_acTL  29
#define PNGTR_fcTL  30
#define PNGTR_fdAT  31
#define PNGTR_vpAg  32
#define PNGTR_orNT  33
#define PNGTR_caNv  34
#define PNGTR_exIf  35

/***************************************************************************************************** 
 ****                                                                                   **** 
 *****************************************************************************************************/
#define PNGTR_DU     1    /* 4, 2  1                            */
#define PNGTR_DD4    2    /* 4                                        */
#define PNGTR_DF4    3    /* 4      / 100000.0 */
#define PNGTR_DF2    4    /* 2      / 100.0    */
#define PNGTR_DW1    5    /* 1                                         */
#define PNGTR_DW2    6    /* 2                                         */
#define PNGTR_DW4    7    /* 4                                                          */
#define PNGTR_DCHAR  8    /*   0                                              */
#define PNGTR_DCHAZ  9    /*                                                       */
#define PNGTR_DTFIX 10    /*   0     ,                   */
#define PNGTR_DW1S  11    /* { "cross-fuse", "diverging-fuse", "Illegal"}                            */
#define PNGTR_DW1R  12    /* 1  , rendering intent                     */
#define PNGTR_DU1W  13    /* 1   ,   8  16          */
#define PNGTR_DU1C  14    /* 1   ,                 */
#define PNGTR_DUS   15    /* { "pixel", "micrometer", "Illegal"}                                     */
#define PNGTR_DUC   16    /* { "unknown", "meter", "radian", "Illegal"}                              */
#define PNGTR_DUP   17    /* { "unknown", "meter", "Illegal"}                                        */
#define PNGTR_DU1N  18    /* 1   ,   0                   */
#define PNGTR_DW1T  19    /* 1  ,                              */
#define PNGTR_DU1I  20    /* 1   , interlace method                */
#define PNGTR_DUH   21    /*                                                              */
#define PNGTR_DCC   22    /* Palette                                                                 */
#define PNGTR_DCA   23    /*                                                             */
#define PNGTR_DCA1  24    /* 1RGBA +                                                          */
#define PNGTR_DCA2  25    /* 2RGBA +                                                          */
#define PNGTR_DC1   26    /* 1RGB                                                                    */
#define PNGTR_DC2   27    /* 2RGB                                                                    */
#define PNGTR_DBYTE 28    /*                                             */
#define PNGTR_DBG   29    /*                                                          */
#define PNGTR_DBZ   30    /*                                              */
#define PNGTR_DBHZ  31    /* Image data                                                */
#define PNGTR_TRASH 32    /*                                                                    */
#define PNGTR_DMAXI 33    /*                                                         */

#define PNGTR_O(IF, T)        do{if(!!(IF)) {ifBLIN_QW0 T; goto out;};}while(0)
#define PNGTR_W(IF, T, EX)    do{if(!!(IF)) {ifBLIN_QW0 T; ex = (EX); goto out;};}while(0)
#define PNGTR_X(IF, T, E, EX) do{if(!!(IF)) {ifBLIN_QX0 T; ex = (EX); errno = (E); goto out;};}while(0)

typedef u_int16_t pngtr_chtype;

typedef struct pngtr_diapa {
    u_int min;
    u_int max;
} pngtr_diapa;

typedef struct pngtr_miapair {
    u_int16_t min;
    u_int16_t max;
} pngtr_miapair;

typedef struct pngtr_miaobj {
    u_int8_t         size;
    int8_t           width;
    pngtr_miapair minimax[0];
} pngtr_miaobj;

typedef struct pngtr_syncolor {
    u_int16_t r;
    u_int16_t g;
    u_int16_t b;
    u_int16_t a;
} pngtr_syncolor;

typedef struct pngtr_color {
    u_char r;
    u_char g;
    u_char b;
} pngtr_color;

typedef struct pngtr_color2 {
    u_char r[2];
    u_char g[2];
    u_char b[2];
} pngtr_color2;

typedef struct pngtr_sPLT1_t {
    pngtr_color  color;
    u_char a;
    u_char freq[2];
} pngtr_sPLT1_t;

typedef struct pngtr_sPLT2_t {
    pngtr_color2 color;
    u_char a[2];
    u_char freq[2];
} pngtr_sPLT2_t;

typedef struct pngtr_hdr {
    u_int8_t ln[4];
    char     nm[4];
    u_char   n[0];
#   define   CRC(H)  ((u_char *)(H) + 8 + recobe_load((H)->ln, -4))
#   define   NEXT(H) ((pngtr_hdr *)((char*)(H) + 12 + recobe_load((H)->ln, -4)))
} pngtr_hdr;

typedef struct pngtr_operand {
    BLIN_flag         flags;
#   define    PNGTR_SELECTD 0x080000    /*                                                     */
#   define    PNGTR_VIRTUAL 0x100000    /*                                       */
    u_int32_t         size;             /*                                             */
    u_char           *val;              /*                                           */
    u_char           *select;           /*                                           */
    u_int32_t         maxindex;         /*  */
} pngtr_operand;

typedef struct pngtr_chunk {
    BLIN_flag         flags;
#   define    PNGTR_SELECTD 0x080000    /*                                                     */
#   define    PNGTR_VIRTUAL 0x100000    /*                                       */
#   define    PNGTR_INFLA   0x200000    /*  infla                                         */
    u_int32_t         csz;              /*                                                */
    pngtr_operand    *operands;         /*                                       */
    u_char           *infla;            /*                                 */
    pngtr_hdr        *vchu;             /*                               */
    size_t            offset;           /*                               */
    size_t            inflen;           /*                          */
    u_int32_t         carg;             /*                                   */
    pngtr_chtype      type;             /*                                                   */
} pngtr_chunk;

typedef struct pngtr_infile {
    mife_descriptor  *md;               /* mife                                         */
    char             *nm;               /*                                         */
    u_int8_t         *cont;             /*                                            */
    ssize_t           size;             /*                                        */
} pngtr_infile;

typedef union pngtr_hdrcolor {
    u_char colortype;
    struct { u_char palette : 1
                  , color   : 1
                  , alfa    : 1
                  , q       : 5
           ;
}; } pngtr_hdrcolor;

typedef struct pngtr_IHDR {
    u_int32_t width;
    u_int32_t height;
    u_char    depth;
    union {
        u_char colortype;
        struct { u_char palette : 1
                      , color   : 1
                      , alfa    : 1
                      , q       : 5
               ;
    }; };
    u_char compress;
    u_char filter;
    u_char interlace;
} pngtr_IHDR;

typedef struct pngtr_convey {
    BLIN_flag         flags;
#   define    PNGTR_INTERLA 0x00007F    /*    Interlace                      */
#   define    PNGTR_SELECTD 0x080000    /* IHDR                                                */
#   define    PNGTR_CHATYPE 0x100000    /*                                         */
#   define    PNGTR_INDEXSL 0x200000    /*                                           */
#   define    PNGTR_OPERSEL 0x400000    /*                                          */
#   define    PNGTR_CHAINSL 0x800000    /*                                             */
    u_int32_t         chunum;           /*                                           */
    pngtr_chunk      *chunks;           /*                                   */
    u_int8_t         *imda;             /*   IDAT                        */
    pngtr_infile      png;              /*   PNG                               */
    pngtr_infile      img;              /*                                   */
    u_int64_t         typemask;         /*                                           */
    u_int32_t         lastchunk;        /*                                    */
    u_int32_t         lastoperand;      /*                                 */
    u_int32_t         lastindex;        /*                                  */
    u_int32_t         nowsz;            /*  imda                                               */
    u_int32_t         chumax;           /*                                   */
    z_stream          strm;             /*   zlib  IDAT                        */
    pngtr_IHDR        IHDR;             /* IHDR PNG                                             */
} pngtr_convey;

typedef struct pngtr_conf {
    BLIN_flag        flags;
                                        /*               */
#   define    PNGTR_PRINTFM 0x000001    /*                           */
#   define    PNGTR_PRINTIM 0x000002    /*                               */
#   define    PNGTR_PRINTFI 0x000004    /*                                    */
#   define    PNGTR_PRINTOP 0x000008    /*                                    */
#   define    PNGTR_PRINTXD 0x000010    /*                              */
#   define    PNGTR_PRINTEX 0x000020    /*                  */
#   define    PNGTR_OUTPAM  0x000040    /*  PAM                                              */
#   define    PNGTR_PRINTFN 0x000100    /*                                           */
#   define    PNGTR_PRINTHD 0x000200    /*                                    */
#   define    PNGTR_NOSIGN  0x000400    /*                             */
#   define    PNGTR_ADDSIGN 0x000800    /*                                          */
#   define    PNGTR_AGRERR  0x001000    /*              */
#   define    PNGTR_PRINTCD 0x002000    /*                                        */
#   define    PNGTR_NOINVAL 0x004000    /*               */
#   define    PNGTR_CFTOCH  0x007FFF    /*  ,     -i                  */
#   define    PNGTR_BINOUT  0x040000    /*                                           */
    u_int32_t        dummy;
    char            *ounm;              /* Path                             */
    pngtr_convey  *convey;            /*                                           */
    FILE            *file;              /*                                               */
    u_int            num;               /*                                      */
    pngtr_IHDR     IHDR;
} pngtr_conf;

typedef struct pngtr_opdef {   /*    */
    const char         *descr;     /*                                               */
    u_int32_t           maxl;      /* OPVAR -                             */
                                   /* OPFIX -                                           */
    u_char              minl;      /*                                      */
    u_char              var;
#   define    PNGTR_OPVAR 0x01     /*                                     */
#   define    PNGTR_OPIDX 0x02     /*  -     (  )   */
#   define    PNGTR_OPFIX 0x04     /*  -                          */
#   define    PNGTR_CHDEP 0x08     /*      Sample depth              */
#   define    PNGTR_CHZIP 0x10     /*      Compression flag          */
#   define    PNGTR_OPNUM 0x20     /*                               */
    u_char              otype;     /*                                                    */
} pngtr_opdef;

typedef struct pngtr_chidef {   /*   */
    u_char              opn;        /*                                            */
    const pngtr_opdef   op[];       /*                                             */
} pngtr_chidef;

typedef struct pngtr_chdef {    /*   */
    const char         *descr;      /*                                              */
    u_char              nm[4];      /*                                         */
    u_char              shift;      /*                                                   */
#   define    PNGTR_CHCOLOR 0x02    /*      IHDR.colortype           */
#   define    PNGTR_CHDEP   0x08    /*      Sample depth             */
#   define    PNGTR_CHZIP   0x10    /*      Compression flag         */
#   define    PNGTR_PNGHEAD 0x80    /*     (IHDR)       */
    const pngtr_chidef**ch;         /*                                            */
} pngtr_chdef;

extern const u_char      pngtr_sg[8];
extern const pngtr_chdef pngtr_chdescr[];
extern const u_int8_t    pngtr_pointsize[7][17];

int           pngtr_dump        __P((pngtr_conf *cf, const void *buf, ssize_t len, FILE *pr));
int           pngtr_minimax     __P((pngtr_conf *cf, pngtr_miaobj *m, void *val));
pngtr_miaobj *pngtr_minit       __P((pngtr_conf *cf, u_int8_t size, int8_t width));
int           pngtr_read        __P((pngtr_conf *cf, const char *o, int mode));
int           pngtr_infla       __P((pngtr_conf *cf, u_int chctr, u_int opctr));
void          pngtr_freeop      __P((pngtr_conf *cf, u_int chctr, u_int opctr));
void          pngtr_freechunk   __P((pngtr_conf *cf, u_int chctr));
int           pngtr_chunksel    __P((pngtr_conf *cf, u_int min, u_int max));
int           pngtr_opsel       __P((pngtr_conf *cf, u_int min, u_int max));
int           pngtr_idxsel      __P((pngtr_conf *cf, u_int min, u_int max));
int           pngtr_print       __P((pngtr_conf *cf, BLIN_flag f, FILE *pr));
int           pngtr_out         __P((pngtr_conf *cf, BLIN_flag f, int fd));
int           pngtr_outpam      __P((pngtr_conf *cf, int fd));
int           pngtr_flimg       __P((pngtr_conf *cf, FILE *pr));
void          pngtr_freefile    __P((pngtr_conf *cf, int mode));
int           pngtr_crechunk    __P((pngtr_conf *cf, pngtr_chtype type));
int           pngtr_vchu        __P((pngtr_conf *cf, pngtr_chunk *chunk));
void          pngtr_freeconvey  __P((pngtr_conf *cf));
int           pngtr_imgfla      __P((pngtr_conf *cf));
int           pngtr_realconvey  __P((pngtr_conf *cf));
int           pngtr_chudelete   __P((pngtr_conf *cf));
int           pngtr_dir         __P((pngtr_conf *cf));
int           pngtr_realimg     __P((pngtr_conf *cf));
